@@ -29,7 +29,7 @@
 char *
 _gdbm_read_entry (GDBM_FILE dbf, int elem_loc)
 {
-  int num_bytes;		/* For seeking and reading. */
+  int rc;
   int key_size;
   int data_size;
   off_t file_pos;
@@ -58,11 +58,13 @@ _gdbm_read_entry (GDBM_FILE dbf, int elem_loc)
 
 
   /* Read into the cache. */
-  file_pos = __lseek (dbf, dbf->bucket->h_table[elem_loc].data_pointer, L_SET);
+  file_pos = __lseek (dbf, dbf->bucket->h_table[elem_loc].data_pointer, 
+                      SEEK_SET);
   if (file_pos != dbf->bucket->h_table[elem_loc].data_pointer)
     _gdbm_fatal (dbf, _("lseek error"));
-  num_bytes = __read (dbf, data_ca->dptr, key_size+data_size);
-  if (num_bytes != key_size+data_size) _gdbm_fatal (dbf, _("read error"));
+  rc = _gdbm_full_read (dbf, data_ca->dptr, key_size+data_size);
+  if (rc)
+    _gdbm_fatal (dbf, gdbm_strerror (rc));
   
   return data_ca->dptr;
 }
