@@ -46,8 +46,8 @@
 # define SEEK_SET        0
 #endif
 
-#ifndef L_SET
-# define L_SET SEEK_SET
+#ifndef O_CLOEXEC
+# define O_CLOEXEC 0
 #endif
 
 /* Default block size.  Some systems do not have blocksize in their
